import logging
from typing import Any

from lxml import etree

from . import action as a
from . import condition as c
from . import constants as const


def _write_base(attribute_name: str, pset_name: str, value_type: str) -> (etree.Element, etree.Element):
    """Schreibt basisinfos die für jede Attributregel identisch sind"""
    rule = etree.Element(const.RULE)
    etree.SubElement(rule, const.IFCTYPE).text = const.ANY
    xml_property = etree.SubElement(rule, const.PROPERTY)
    etree.SubElement(xml_property, const.NAME).text = attribute_name
    etree.SubElement(xml_property, const.PSETNAME).text = pset_name
    etree.SubElement(xml_property, const.TYPE).text = const.PROPERTYSET
    etree.SubElement(xml_property, const.VALUETYPE).text = value_type
    etree.SubElement(xml_property, const.UNIT).text = const.NONE
    xml_condition = etree.SubElement(rule, const.CONDITION)
    return rule, xml_condition


def _generate_rule(attribute_name: str, property_set_name: str,
                   value: Any, datatype: str, condition_type: str, action_type: str) -> etree.Element:
    xml_rule, xml_condition = _write_base(attribute_name, property_set_name, datatype)
    etree.SubElement(xml_condition, const.TYPE).text = condition_type
    etree.SubElement(xml_condition, const.VALUE).text = str(value)
    etree.SubElement(etree.SubElement(xml_rule, a.ACTION), const.TYPE).text = action_type
    return xml_rule


def add_if_not_existing(attribute_name: str, pset_name: str, data_type: str) -> list[etree.Element]:
    if data_type == c.DOUBLE:
        c_type = c.NUMERIC_UNDEF
    elif data_type == c.BOOL:
        c_type = c.BOOL_UNDEF
    elif data_type == c.STRING:
        c_type = c.STRING_UNDEF
    else:
        logging.error(f"Datatype {data_type} unknown. Skip Rule")
        return list()
    return [_generate_rule(attribute_name, pset_name, data_type, data_type, c_type,
                           a.ADD)]


def remove_if_not_in_string_list(attribute_name: str, pset_name: str, value_list) -> list[etree.Element]:
    return [_generate_rule(attribute_name, pset_name, ",".join(value_list), c.STRING, c.NOR, a.REMOVE)]


def add_if_not_in_string_list(attribute_name: str, pset_name: str, value_list) -> list[etree.Element]:
    return [_generate_rule(attribute_name, pset_name, ",".join(value_list), c.STRING, c.NOR, a.ADD)]


def remove_if_in_string_list(attribute_name: str, pset_name: str, value_list) -> list[etree.Element]:
    return [_generate_rule(attribute_name, pset_name, ",".join(value_list), c.STRING, c.OR, a.REMOVE)]


def add_if_in_string_list(attribute_name: str, pset_name: str, value_list) -> list[etree.Element]:
    return [_generate_rule(attribute_name, pset_name, ",".join(value_list), c.STRING, c.OR, a.ADD)]


def add_if_outside_of_range(attribute_name: str, pset_name: str, min_value: float, max_value: float) -> \
        list[etree.Element]:
    r1 = _generate_rule(attribute_name, pset_name, min_value, c.DOUBLE, c.LESS, a.ADD)
    r2 = _generate_rule(attribute_name, pset_name, max_value, c.DOUBLE, c.GREATER, a.ADD)
    return [r1, r2]


def add_if_in_range(attribute_name: str, pset_name: str, min_value: float, max_value: float) -> list[etree.Element]:
    r1 = _generate_rule(attribute_name, pset_name, min_value, c.DOUBLE, c.GREATER, a.AND)
    r2 = _generate_rule(attribute_name, pset_name, max_value, c.DOUBLE, c.LESS, a.ADD)
    return [r1, r2]


def numeric_list(attribute_name, pset_name, value_list) -> list[etree.Element]:
    """Schreibt Regel die kontrolliert ob ein Zahlenwert in einer Liste aus Zahlen vorkommt"""
    rule_list: list[etree.Element] = list()
    rule_list += add_if_not_existing(attribute_name, pset_name, c.DOUBLE)
    for index, value in enumerate(value_list):
        if index == len(value_list) - 1:
            xml_rule = _generate_rule(attribute_name, pset_name, value, c.DOUBLE, c.NOTEQ, a.ADD)
        else:
            xml_rule = _generate_rule(attribute_name, pset_name, value, c.DOUBLE, c.NOTEQ, a.AND)
        rule_list.append(xml_rule)
    return rule_list


def merge_list(range_list, start_index=0):
    for i in range(start_index, len(range_list) - 1):
        if range_list[i][1] > range_list[i + 1][0]:
            new_start = range_list[i][0]
            new_end = max(range_list[i + 1][1], range_list[i][1])
            range_list[i] = [new_start, new_end]
            del range_list[i + 1]
            return merge_list(range_list.copy(), start_index=i)
    return range_list


def numeric_range(attribute_name: str, property_set_name: str,
                  value_range_list: list[tuple[float, float]]) -> list[etree.Element]:
    """Schreibt eine Regel die kontrolliert ob ein Zahlenwert in einem Wertebereich vorkommt"""

    if not value_range_list:
        logging.error(f"Empty Value list at {property_set_name}:{attribute_name}")
        return list()

    sorted_range_list = sorted([[min(v1, v2), max(v1, v2)] for [v1, v2] in value_range_list])
    sorted_range_list = merge_list(sorted_range_list)

    minimal_value = sorted_range_list[0][0]
    maximal_value = sorted_range_list[-1][1]

    rule_list = list()
    rule_list += add_if_not_existing(attribute_name, property_set_name, c.DOUBLE)
    rule_list += add_if_outside_of_range(attribute_name, property_set_name, minimal_value, maximal_value)

    for [_, v1_max], [v2_min, _] in zip(sorted_range_list, sorted_range_list[1:]):
        rule_list += add_if_in_range(attribute_name, property_set_name, v1_max, v2_min)
    return rule_list


def remove_if_not_exist(attribute_name: str, property_set_name: str) -> list[etree.Element]:
    return _generate_rule(attribute_name, property_set_name, "", c.STRING, c.ISNOT, a.REMOVE)



from __future__ import annotations
import datetime
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element
from lxml import etree

from src.SOMcreator import classes
output_date_time = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")


def handle_header(project: classes.Project, export_format: str) -> Element:
    ET.register_namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")
    xml_header = etree.Element(f'{{http://www.w3.org/2001/XMLSchema-instance}}{export_format}')
    xml_header.set("user", str(project.author))
    xml_header.set("date", str(output_date_time))
    xml_header.set("version", "3.0.1")  # TODO: Desite version hinzufügen
    return xml_header



from src.SOMcreator.constants import value_constants

STRING = "StringValue"
DOUBLE = "DoubleValue"
BOOL = "BoolValue"
LOGICAL = "LogicalValue"







#String Conditions
ISNOT = "IsNot"
IS = "Is"
CONTAINS = "Contains"
DOESNOTCONTAIN = "DoesNotContain"
STARTSWITH ="StartsWith"
ENDSWITH = "EndsWith"
OR = "Or"
NOR = "Nor"
STRING_DEF = "StringIsDefined"
STRING_UNDEF = "StringIsNotDefined"

#NumberConditions
EQUALS = "Equals"
NOTEQ = "NotEquals"
LESS = "Less"
GREATER = "Greater"
LESSEQUAL = "LessEqual"
GREATEREQUAL = "GreaterEqual"
NUMERIC_DEF = "NumericIsDefined"
NUMERIC_UNDEF = "NumericIsNotDefined"

#BoolConditions
BOOLTRUE = "BooleanTrue"
BOOLFALSE = "BooleanFalse"
BOOL_DEF = "BooleanIsDefined"
BOOL_UNDEF = "BooleanIsNotDefined"

#LogicContions
LOGICTRUE = "LogicalTrue"
LOGICFALSE = "LogicalFalse"
LOGICUNKNOW = "LogicalUnknown"
LOGICAL_DEF = "LogicalIsDefined"
LOGICAL_UNDEF = "LogicalIsNotDefined"


DATATYPE_DICT = {value_constants.XS_STRING: STRING,
                 value_constants.XS_BOOL: BOOL,
                 value_constants.XS_INT: DOUBLE,
                 value_constants.XS_DOUBLE: DOUBLE,
                 value_constants.XS_LONG: DOUBLE}


from __future__ import annotations

import json
import logging
import os

import jinja2

from . import classes
from .Template import MAPPING_TEMPLATE, HOME_DIR
from .constants import json_constants


def string_to_bool(text: str) -> bool | None:
    if text == str(True):
        return True
    elif text == str(False):
        return False
    else:
        return None


def create_mapping_script(project: classes.Project, pset_name: str, path: str):
    attrib_dict = dict()
    obj: classes.Object
    for obj in project.objects:
        klass = obj.ident_attrib.value[0]
        obj_dict = dict()
        for pset in obj.property_sets:
            pset_dict = dict()
            for attribute in pset.attributes:
                name = attribute.name
                data_format = attribute.data_type
                pset_dict[name] = data_format
            obj_dict[pset.name] = pset_dict
        attrib_dict[klass] = obj_dict
    file_loader = jinja2.FileSystemLoader(HOME_DIR)
    env = jinja2.Environment(loader=file_loader)
    env.trim_blocks = True
    env.lstrip_blocks = True

    template = env.get_template(MAPPING_TEMPLATE)
    code = template.render(attribute_dict=attrib_dict, pset_name=pset_name)
    with open(path, "w") as file:
        file.write(code)
    pass


def export_json(project: classes.Project, path: str) -> dict:
    def create_project_data(project_dict: dict) -> None:
        project_dict[json_constants.NAME] = project.name
        project_dict[json_constants.AUTHOR] = project.author
        project_dict[json_constants.VERSION] = project.version
        project_dict[json_constants.AGGREGATION_ATTRIBUTE] = project.aggregation_attribute
        project_dict[json_constants.AGGREGATION_PSET] = project.aggregation_pset
        project_dict[json_constants.CURRENT_PR0JECT_PHASE] = project.current_project_phase
        project_dict[json_constants.PROJECT_PHASES] = project.get_project_phase_list()

    def fill_basics(entity_dict, entity):
        """function gets called from all Entities"""
        entity_dict[json_constants.NAME] = entity.name
        entity_dict[json_constants.OPTIONAL] = entity.optional
        entity_dict[json_constants.PROJECT_PHASES] = entity.get_project_phase_dict()
        parent = None if entity.parent is None else entity.parent.uuid
        entity_dict[json_constants.PARENT] = parent
        entity_dict[json_constants.DESCRIPTION] = entity.description

    def create_attribute_entry(attributes_dict, attribute):
        attribute_dict = dict()
        fill_basics(attribute_dict, attribute)
        attribute_dict[json_constants.DATA_TYPE] = attribute.data_type
        attribute_dict[json_constants.VALUE_TYPE] = attribute.value_type
        attribute_dict[json_constants.CHILD_INHERITS_VALUE] = attribute.child_inherits_values
        attribute_dict[json_constants.REVIT_MAPPING] = attribute.revit_name
        attribute_dict[json_constants.VALUE] = attribute.value
        attributes_dict[attribute.uuid] = attribute_dict

    def create_pset_entry(psets_dict: dict, pset: classes.PropertySet):
        pset_dict = dict()
        fill_basics(pset_dict, pset)
        attributes_dict = dict()
        for attribute in pset.get_all_attributes():
            create_attribute_entry(attributes_dict, attribute)
        pset_dict[json_constants.ATTRIBUTES] = attributes_dict
        psets_dict[pset.uuid] = pset_dict

    def create_object_entry(objects_dict: dict, object: classes.Object):

        object_dict = dict()
        fill_basics(object_dict, object)

        if isinstance(obj.ifc_mapping, set):
            object_dict[json_constants.IFC_MAPPINGS] = list(object.ifc_mapping)
        else:
            object_dict[json_constants.IFC_MAPPINGS] = object.ifc_mapping

        psets_dict = dict()
        for pset in object.get_all_property_sets():
            create_pset_entry(psets_dict, pset)

        object_dict[json_constants.PROPERTY_SETS] = psets_dict
        object_dict[json_constants.ABBREVIATION] = object.abbreviation

        objects_dict[obj.uuid] = object_dict
        if isinstance(object.ident_attrib, classes.Attribute):
            object_dict[json_constants.IDENT_ATTRIBUTE] = object.ident_attrib.uuid
        else:
            object_dict[json_constants.IDENT_ATTRIBUTE] = object.ident_attrib

    def create_aggregation_entry(aggregations_dict, aggregation: classes.Aggregation):
        aggregation_dict = dict()
        fill_basics(aggregation_dict, aggregation)
        aggregation_dict[json_constants.OBJECT] = aggregation.object.uuid
        if aggregation.parent is not None:
            aggregation_dict[json_constants.PARENT] = aggregation.parent.uuid
        else:
            aggregation_dict[json_constants.PARENT] = aggregation.parent
        aggregation_dict[json_constants.CONNECTION] = aggregation.parent_connection
        aggregations_dict[aggregation.uuid] = aggregation_dict

    main_dict = dict()
    main_dict[json_constants.PROJECT] = dict()
    create_project_data(main_dict[json_constants.PROJECT])

    predef_pset_dict = dict()
    predefined_psets = project.get_predefined_psets()
    for entity in sorted(predefined_psets, key=lambda x: x.uuid):
        create_pset_entry(predef_pset_dict, entity)
    main_dict[json_constants.PREDEFINED_PSETS] = predef_pset_dict

    objects_dict = dict()
    for obj in project.get_all_objects():
        create_object_entry(objects_dict, obj)

    main_dict[json_constants.OBJECTS] = objects_dict

    aggregations_dict = dict()
    for aggregation in project.get_all_aggregations():
        create_aggregation_entry(aggregations_dict, aggregation)
    main_dict[json_constants.AGGREGATIONS] = aggregations_dict
    with open(path, "w") as file:
        json.dump(main_dict, file, indent=2)

    return main_dict


def import_json(project: classes.Project, path: str):
    if not os.path.isfile(path):
        return
    parent_dict = {}
    aggregation_parent_dict: dict[classes.Aggregation, (str, int)] = dict()

    with open(path, "r") as file:
        main_dict: dict = json.load(file)

    def load_project_data(project_dict: dict):
        project.name = project_dict.get(json_constants.NAME)
        project.author = project_dict.get(json_constants.AUTHOR)
        project.version = project_dict.get(json_constants.VERSION)

        pset = project_dict.get(json_constants.AGGREGATION_PSET)
        attribute = project_dict.get(json_constants.AGGREGATION_ATTRIBUTE)
        current_project_phase = project_dict.get(json_constants.CURRENT_PR0JECT_PHASE)
        project_phases = project_dict.get(json_constants.PROJECT_PHASES)

        if pset is not None:
            project.aggregation_pset = pset
        if attribute is not None:
            project.aggregation_attribute = attribute
        if project_phases is not None and isinstance(project_phases, list):
            project._project_phases = project_phases

        if current_project_phase is not None:
            project.current_project_phase = current_project_phase
        elif project.get_project_phase_list():
            project.current_project_phase = project.get_project_phase_list()[0]

    def load_basics(element_dict: dict) -> (str, str, str):
        name = element_dict[json_constants.NAME]
        description = element_dict[json_constants.DESCRIPTION]
        optional = element_dict[json_constants.OPTIONAL]
        parent = element_dict[json_constants.PARENT]
        project_phases = element_dict.get(json_constants.PROJECT_PHASES)
        if not isinstance(project_phases, dict):
            project_phases = None
        return name, description, optional, parent, project_phases

    def load_object(object_dict, identifier):
        name, description, optional, parent, project_phases = load_basics(object_dict)
        ifc_mapping = object_dict[json_constants.IFC_MAPPINGS]
        if isinstance(ifc_mapping, list):
            ifc_mapping = set(ifc_mapping)

        abbreviation = object_dict.get(json_constants.ABBREVIATION)

        obj = classes.Object(name=name, ident_attrib=None, uuid=identifier, ifc_mapping=ifc_mapping,
                             description=description, optional=optional, abbreviation=abbreviation, project=project,
                             project_phases=project_phases)
        property_sets_dict = object_dict[json_constants.PROPERTY_SETS]
        for ident, pset_dict in property_sets_dict.items():
            load_pset(pset_dict, ident, obj)
        ident_attrib_id = object_dict[json_constants.IDENT_ATTRIBUTE]
        ident_attrib = classes.get_element_by_uuid(ident_attrib_id)
        obj.ident_attrib = ident_attrib
        parent_dict[obj] = parent

    def load_pset(pset_dict: dict, identifier: str, obj: classes.Object | None) -> None:
        name, description, optional, parent, project_phases = load_basics(pset_dict)
        pset = classes.PropertySet(name=name, obj=obj, uuid=identifier, description=description, optional=optional,
                                   project=project, project_phases=project_phases)
        attributes_dict = pset_dict[json_constants.ATTRIBUTES]
        for ident, attribute_dict in attributes_dict.items():
            load_attribute(attribute_dict, ident, pset)
        parent_dict[pset] = parent

    def load_attribute(attribute_dict: dict, identifier: str, property_set: classes.PropertySet) -> None:
        name, description, optional, parent, project_phases = load_basics(attribute_dict)
        value = attribute_dict[json_constants.VALUE]
        value_type = attribute_dict[json_constants.VALUE_TYPE]
        data_type = attribute_dict[json_constants.DATA_TYPE]
        child_inherits_value = attribute_dict[json_constants.CHILD_INHERITS_VALUE]
        revit_mapping = attribute_dict[json_constants.REVIT_MAPPING]
        attribute = classes.Attribute(property_set=property_set, name=name, value=value, value_type=value_type,
                                      data_type=data_type,
                                      child_inherits_values=child_inherits_value, uuid=identifier,
                                      description=description, optional=optional, revit_mapping=revit_mapping,
                                      project=project, project_phases=project_phases)
        parent_dict[attribute] = parent

    def load_dict(dict_name: str) -> dict | None:
        return_dict = main_dict.get(dict_name)
        if return_dict is None:
            logging.error(f"loading Error: {dict_name} doesn't exist!")
            return {}
        return return_dict

    def load_parents():
        def find_parent(element):
            print(f"gesucht: {element.name}")
            for test_el,uuid in parent_dict.items():
                if type(test_el) != type(element):
                    continue
                if uuid not in uuid_dict:
                    continue
                if test_el == element:
                    continue
                if test_el.parent is not None:
                    continue

                if test_el.name != element.name:
                    continue

                if isinstance(test_el,classes.Attribute):
                    if test_el.value == element.value:
                        return uuid
                        continue
                return uuid


        uuid_dict = classes.get_uuid_dict()
        for element, uuid in parent_dict.items():
            if uuid is not None:
                if uuid not in uuid_dict:
                    uuid = find_parent(element)
                    print(f"neue uuid: {uuid}")
                if uuid is None:
                    continue
                uuid_dict[uuid].add_child(element)

    def load_aggregation(aggregation_dict: dict, identifier: str, ):
        name, description, optional, parent, project_phases = load_basics(aggregation_dict)
        object_uuid = aggregation_dict[json_constants.OBJECT]
        obj = classes.get_element_by_uuid(object_uuid)
        parent_connection = aggregation_dict[json_constants.CONNECTION]
        aggregation = classes.Aggregation(obj, parent_connection, identifier, description, optional, project_phases)
        aggregation_parent_dict[aggregation] = (parent, parent_connection)

    def build_aggregation_structure():
        for aggregation, (uuid, connection_type) in aggregation_parent_dict.items():
            parent = classes.get_element_by_uuid(uuid)
            if parent is not None:
                parent.add_child(aggregation, connection_type)

    def fill_missing_project_phases():
        phases = project.get_project_phase_list()
        for item in project.get_all_hirarchy_items():
            phase_dict = item.get_project_phase_dict()
            for phase in phases:
                if phase not in phase_dict:
                    item.add_project_phase(phase, True)

    project_dict = main_dict.get(json_constants.PROJECT)
    if project_dict is None:
        logging.warning(f"{json_constants.PROJECT}-dict doesn't exist unable to load Author, Name and Version")
    else:
        load_project_data(project_dict)

    predefined_psets_dict = load_dict(json_constants.PREDEFINED_PSETS)

    for ident, pset_dict in predefined_psets_dict.items():
        load_pset(pset_dict, ident, None)

    objects_dict = load_dict(json_constants.OBJECTS)
    for ident, object_dict in objects_dict.items():
        load_object(object_dict, ident)

    aggregations_dict = load_dict(json_constants.AGGREGATIONS)
    for ident, aggreg_dict in aggregations_dict.items():
        load_aggregation(aggreg_dict, ident)

    load_parents()
    build_aggregation_structure()
    fill_missing_project_phases()
    return main_dict



from __future__ import annotations

import datetime
import logging
import os
import uuid

from lxml import etree

from constants import value_constants
from src.SOMcreator import classes
from . import condition as c
from . import constants as const
from . import rule


def _write_header(xml_header: etree.Element) -> None:
    etree.SubElement(xml_header, const.VERSION).text = "6"
    etree.SubElement(xml_header, const.APPVER).text = "Win - Version: 6.8 (build 6.8.26.0)"


def _write_smartview(property_set: classes.PropertySet, author: str) -> etree.Element:
    def write_smartview_basics():
        sv = etree.Element(const.SVIEW)
        etree.SubElement(sv, const.TITLE).text = property_set.name
        etree.SubElement(sv, const.DESCRIPTION).text = f"Checks {property_set.name} for correct Values"
        etree.SubElement(sv, const.CREATOR).text = "christoph.mellueh@deutschebahn.com"
        etree.SubElement(sv, const.CREATIONDATE).text = str(datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
        etree.SubElement(sv, const.MODIFIER).text = author
        etree.SubElement(sv, const.MODIFICATIONDATE).text = str(datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
        etree.SubElement(sv, const.GUID).text = str(uuid.uuid4())
        return sv

    xml_smart_view = write_smartview_basics()
    xml_rules = etree.SubElement(xml_smart_view, const.RULES)
    pset_name = property_set.name
    ident_attrib = property_set.object.ident_attrib
    rule_list: list[etree.Element] = list()

    for attribute in property_set.attributes:
        if attribute == ident_attrib:
            continue
        if attribute.data_type in (value_constants.XS_INT, value_constants.XS_LONG, value_constants.XS_DOUBLE):
            if not attribute.value:
                rule_list += rule.add_if_not_existing(attribute.name, pset_name, c.DATATYPE_DICT[attribute.data_type])
            elif attribute.value_type == value_constants.LIST:
                rule_list += rule.numeric_list(attribute.name, pset_name, attribute.value)
            elif attribute.value_type == value_constants.RANGE:
                rule_list += rule.numeric_range(attribute.name, pset_name, attribute.value)
            else:
                logging.error(f"No Function defined for {attribute.name} ({attribute.value_type}x{attribute.data_type}")

        elif attribute.data_type == value_constants.XS_STRING:
            if attribute.value_type == value_constants.FORMAT:
                rule_list += rule.add_if_not_existing(attribute.name, pset_name, c.DATATYPE_DICT[attribute.data_type])
                continue

            if attribute.value:
                rule_list += rule.add_if_not_in_string_list(attribute.name, pset_name, attribute.value)
            else:
                rule_list += rule.add_if_not_existing(attribute.name, pset_name, c.DATATYPE_DICT[attribute.data_type])

        elif attribute.data_type == value_constants.XS_BOOL:
            rule_list += rule.add_if_not_existing(attribute.name, pset_name, c.DATATYPE_DICT[attribute.data_type])
        else:
            logging.error(f"No Function defined for {attribute.name} ({attribute.value_type}x{attribute.data_type}")

    rule_list += rule.remove_if_not_in_string_list(ident_attrib.name, ident_attrib.property_set.name,
                                                   ident_attrib.value)
    for xml_rule in rule_list:
        xml_rules.append(xml_rule)
    return xml_smart_view


def _write_smartviewset(obj: classes.Object, author: str) -> etree.Element:
    smartview_set = etree.Element(const.SMVSET)
    etree.SubElement(smartview_set, const.TITLE).text = obj.name
    etree.SubElement(smartview_set, const.DESCRIPTION).text = "generated by SOMcreator"
    etree.SubElement(smartview_set, const.GUID).text = str(uuid.uuid4())
    etree.SubElement(smartview_set, const.MODIFICATIONDATE).text = str(
        datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
    smartviews = etree.SubElement(smartview_set, const.SVIEWS)

    for property_set in list(obj.property_sets):
        smartviews.append(_write_smartview(property_set, author))
    return smartview_set


def _write_smartviewsets(object_list: list[classes.Object], author: str) -> etree.Element:
    smartviewsets = etree.Element(const.SMVSETS)
    for obj in object_list:
        if obj.is_concept:
            continue
        smartviewsets.append(_write_smartviewset(obj, author))
    return smartviewsets


def main(object_list: list[classes.Object], save_path: os.PathLike | str, author="", ) -> None:
    header = etree.Element(const.BCSVF)
    _write_header(header)
    svs = _write_smartviewsets(object_list, author)

    with open(save_path, "wb") as file:
        file.write(etree.tostring(header, pretty_print=True))
        file.write(etree.tostring(svs, pretty_print=True))



from .classes import PropertySet,Object,Project,Attribute
from .external_software import desite,excel,allplan,revit,vestra,card1
from .constants import json_constants,value_constants
__version__ = "1.3.0"



from __future__ import annotations

import logging
import os

import ifcopenshell
from ifcopenshell.util import element

from .. import classes

FILE_SPLIT = "; "
ELEMENT = "Element"
GROUP = "SubGroup"
IFC_REP = "IfcRep"
DESCRIPTION = "automatisch erzeugt"
NAME = "bauteilName"


def get_ifc_el_info(entity: ifcopenshell.entity_instance, attribute_bundle: tuple[str, str, str, str, str, str]) \
        -> tuple[str | None, str | None, str | None]:
    def check_for_existence(pset_name: str, attribute_name: str):
        pset_dict = psets.get(pset_name)
        if pset_dict is None:
            logging.warning(
                f"Für Entität {entity.GlobalId} konnte keine Gruppe erstellt werden da das Propertyset "
                f"'{pset_name}' nicht gefunden werden konnte")
            return False, None
        value = pset_dict.get(attribute_name)

        if value is None:
            logging.warning(
                f"Für Entität {entity.GlobalId} konnte keine Gruppe erstellt werden da das Attribut "
                f"'{pset_name}:{attribute_name}' nicht gefunden werden konnte")
            return False, None
        return True, value

    main_pset, main_attribute, group_pset, group_attribute, identity_pset, identity_attribute = attribute_bundle
    identity_attribute: str
    """-> Bauteilklassifikation, idGruppe"""
    psets = element.get_psets(entity)

    check_1, attrib = check_for_existence(main_pset, main_attribute)
    check_2, gruppe = check_for_existence(group_pset, group_attribute)
    check_3, identity = check_for_existence(identity_pset, identity_attribute)

    if all((check_1, check_2, check_3)):
        return attrib, gruppe, identity
    else:
        return None, None, None


def create_structure_dict(ifc_file: ifcopenshell.file, project: classes.Project,
                          attribute_bundle: tuple[str, str, str, str, str, str]) -> dict:
    """Iterate over all Entities, build the targeted Datastructure"""
    targeted_group_structure = {GROUP: {}, ELEMENT: {}, IFC_REP: None}
    bk_dict = {obj.ident_value: obj for obj in project.objects}

    for index, el in enumerate(list(ifc_file.by_type("IfcElement"))):
        attrib, gruppe, identity = get_ifc_el_info(el, attribute_bundle)
        if attrib is None or gruppe is None:
            continue

        parts = gruppe.upper().split("_")
        focus_dict = targeted_group_structure
        for part in parts:
            if part not in focus_dict[GROUP]:
                focus_dict[GROUP][part] = {GROUP: {}, ELEMENT: list(), IFC_REP: None}
            focus_dict = focus_dict[GROUP][part]

        obj: classes.Object = bk_dict.get(attrib)
        abbrev = obj.abbreviation
        if abbrev.upper() not in focus_dict[GROUP]:
            focus_dict[GROUP][abbrev] = {GROUP: {}, ELEMENT: list(), IFC_REP: None}
        focus_dict[GROUP][abbrev][ELEMENT].append(el)
    return targeted_group_structure


def fill_existing_groups(ifc_file: ifcopenshell.file, structure_dict: dict,
                         attribute_bundle: tuple[str, str, str, str, str, str]) -> None:
    """Take existing Groups by identity attribute and sort them into the Datastructure"""

    for group in ifc_file.by_type("IfcGroup"):
        attrib, gruppe, identity = get_ifc_el_info(group, attribute_bundle)
        if identity is None:
            continue
        parts = identity.upper().split("_")
        focus_dict = structure_dict
        skip = False
        for part in parts:
            if part in focus_dict[GROUP] and not skip:
                focus_dict = focus_dict[GROUP][part]
            else:
                skip = True
        if skip:
            continue

        focus_dict[IFC_REP] = group


def create_aggregation_structure(ifc_file: ifcopenshell.file, structure: dict, id_gruppe: list[str],
                                 parent_group, is_sammler: bool, attribute_bundle, owner_history, kuerzel_dict,
                                 fill_with_empty_values: bool, obj=None):
    """Take targeted group structure and build it in IFC-File"""

    main_pset, main_attribute, group_pset, group_attribute, identity_pset, identity_attribute = attribute_bundle

    def create_ifc_property(property_name: str, value: str):
        if value is None:
            prop = ifc_file.create_entity("IfcPropertySingleValue", property_name, DESCRIPTION,
                                          None, None)
        else:
            prop = ifc_file.create_entity("IfcPropertySingleValue", property_name, DESCRIPTION,
                                          ifc_file.create_entity("IfcLabel", value), None)
        return prop

    def create_ifc_pset(pset_name, attribute_dict: dict[str, str],
                        relating_entity: ifcopenshell.entity_instance | None = None):
        properties = list()
        for property_name, value in attribute_dict.items():
            properties.append(create_ifc_property(property_name, value))
        pset = ifc_file.create_entity("IfcPropertySet", ifcopenshell.guid.new(), owner_history, pset_name,
                                      DESCRIPTION, properties)
        if relating_entity is not None:
            ifc_file.create_entity("IfcRelDefinesByProperties", ifcopenshell.guid.new(), owner_history,
                                   f"Relationship {pset_name}", DESCRIPTION, [relating_entity], pset)
        return pset

    def fill_group_with_empty_values(ifc_group: ifcopenshell.entity_instance, group_obj: classes.Object, identity):
        for pset in group_obj.property_sets:
            attributes = {attribute.name: None for attribute in pset.attributes}
            if pset.name == main_pset:
                attributes[main_attribute] = group_obj.ident_value
                attributes[NAME] = group_obj.name
                attributes[identity_attribute] = "_".join(identity)

            if pset.name == group_pset:
                attributes[group_attribute] = "_".join(identity[:-2])
            create_ifc_pset(pset.name, attributes, ifc_group)

    def create_ifc_group(group_obj: classes.Object, group_name: str, identity: list[str],
                         parent: ifcopenshell.entity_instance = None) -> ifcopenshell.entity_instance:
        logging.info(f"create_new_group: {group_name}")
        ifc_group = ifc_file.create_entity("IfcGroup", ifcopenshell.guid.new(), owner_history, group_name,
                                           DESCRIPTION)
        ifc_file.create_entity("IfcRelAssignsToGroup", ifcopenshell.guid.new(), owner_history, group_obj.name,
                               DESCRIPTION, [], None, ifc_group)

        if not is_sammler and fill_with_empty_values:
            fill_group_with_empty_values(ifc_group, group_obj, identity)
        else:
            attributes = dict()
            attributes[main_attribute] = group_obj.ident_value
            attributes[NAME] = group_obj.name
            attributes[identity_attribute] = "_".join(identity)
            create_ifc_pset(main_pset, attributes, ifc_group)
            attributes = dict()
            if is_sammler:
                attributes[group_attribute] = "_".join(identity[:-1])
            else:
                attributes[group_attribute] = "_".join(identity[:-2])
            if group_pset in [pset.name for pset in group_obj.property_sets]:
                create_ifc_pset(group_pset, attributes, ifc_group)

        if parent is not None:
            for is_grouped_relation_ship in parent.IsGroupedBy:
                is_grouped_relation_ship[4] = list(is_grouped_relation_ship[4]) + [ifc_group]
                return ifc_group
        return ifc_group

    for abbreviation in structure[GROUP]:
        ifc_rep = structure[GROUP][abbreviation][IFC_REP]
        new_id_gruppe = id_gruppe + [abbreviation]
        if is_sammler:
            obj = kuerzel_dict.get(abbreviation.upper())
            if obj is None:
                continue
            name = obj.name
        else:
            name = f"{obj.name}_{abbreviation}"
        if ifc_rep is None:
            group = create_ifc_group(obj, name, new_id_gruppe, parent_group)
            structure[GROUP][abbreviation][IFC_REP] = group
        else:
            group = ifc_rep
        create_aggregation_structure(ifc_file, structure[GROUP][abbreviation], new_id_gruppe, group,
                                     not is_sammler, attribute_bundle, owner_history, kuerzel_dict,
                                     fill_with_empty_values, obj)

    if not is_sammler:
        for relation_ship in parent_group.IsGroupedBy:
            relation_ship[4] = list(set(relation_ship[4]).union(set(structure[ELEMENT])))
        return


def main(ifc_path: os.PathLike | str, export_path: os.PathLike | str, project: classes.Project,
         main_pset: str, main_attribute: str,
         group_pset: str, group_attribute: str,
         identity_pset: str, identity_attribute: str, fill_with_empty_values=False):
    ifc_file = ifcopenshell.open(ifc_path)
    owner_history = list(ifc_file.by_type("IfcOwnerHistory"))[0]
    attribute_bundle = (main_pset, main_attribute, group_pset, group_attribute, identity_pset, identity_attribute)

    targeted_group_structure = create_structure_dict(ifc_file, project, attribute_bundle)
    fill_existing_groups(ifc_file, targeted_group_structure, attribute_bundle)
    kuerzel_dict = {obj.abbreviation.upper(): obj for obj in project.objects}
    create_aggregation_structure(ifc_file, targeted_group_structure, [], None, True, attribute_bundle, owner_history,
                                 kuerzel_dict, fill_with_empty_values, None)
    ifc_file.write(export_path)



from __future__ import annotations
import os
import jinja2
from lxml import etree
from Template import HOME_DIR, BOOKMARK_TEMPLATE
from src.SOMcreator import classes


def _handle_bookmark_list(proj: classes.Project) -> etree.ElementTree:
    xml_bookmarks = etree.Element("bookmarks")
    xml_bookmarks.set("xmlnsxsi", "http://www.w3.org/2001/XMLSchema-instance")
    xml_bookmark_list = etree.SubElement(xml_bookmarks, "cBookmarkList")

    obj: classes.Object
    for obj in sorted(proj.objects, key=lambda o: o.ident_value):
        xml_bookmark = etree.SubElement(xml_bookmark_list, "cBookmark")
        xml_bookmark.set("ID", str(obj.uuid))

        if isinstance(obj.ident_attrib, classes.Attribute):
            xml_bookmark.set("name", str(obj.ident_attrib.value[0]))

        xml_bookmark.set("bkmType", "2")
        xml_col = etree.SubElement(xml_bookmark, "col")
        xml_col.set("v", "Type##xs:string")

        attribute = obj.ident_attrib
        xml_col = etree.SubElement(xml_bookmark, "col")
        text = f"{attribute.property_set.name}:{attribute.name}##{attribute.data_type}"
        xml_col.set("v", text)

        for property_set in obj.property_sets:
            for attribute in property_set.attributes:
                if attribute != obj.ident_attrib:
                    xml_col = etree.SubElement(xml_bookmark, "col")
                    text = f"{property_set.name}:{attribute.name}##{attribute.data_type}"
                    xml_col.set("v", text)
    return etree.ElementTree(xml_bookmarks)


def _get_attribute_dict(proj: classes.Project) -> dict[str, str]:
    attribute_dict = {}
    for obj in proj.objects:
        for property_set in obj.property_sets:
            for attribute in property_set.attributes:
                attribute_dict[f"{property_set.name}:{attribute.name}"] = attribute.data_type

    return attribute_dict


def export_bookmarks(proj: classes.Project, path: str) -> None:
    if not os.path.isdir(path):
        return

    with open(os.path.join(path, "bookmarks.bkxml"), "wb") as f:
        tree = _handle_bookmark_list(proj)
        tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")

    attrib_dict = _get_attribute_dict(proj)
    file_loader = jinja2.FileSystemLoader(HOME_DIR)
    env = jinja2.Environment(loader=file_loader)
    env.trim_blocks = True
    env.lstrip_blocks = True
    template = env.get_template(BOOKMARK_TEMPLATE)
    code = template.render(attribute_dict=attrib_dict)
    with open(os.path.join(path, "bookmark_script.js"), "w") as f:
        f.write(code)



NONE = "None"
NOR = "Nor"
ANY = "Any"
ADD = "Add"
REMOVE = "Remove"
AND = "And..."
REMOVE_OTHERS = "RemoveOthers"
SETCOLORED = "SetTransparent"
SETOPAQUE = "SetOpaque"
SETAC = "SetAutoColored"
ADDSETCOL = "AddSetColored"
ADDSETTRANS = "AddSetTransparent"
ACTION = "ACTION"
TYPE = "ACTION.TYPE"
R = "ACTION.R"
G = "ACTION.G"
B = "ACTION.B"



VALUE = "Value"
FORMAT = "Format"
RANGE = "Range"
LIST = "List"
XS_STRING = "xs:string"
XS_DOUBLE = "xs:double"
XS_BOOL = "xs:boolean"
XS_LONG = "xs:long"
DATATYPE_NUMBER = "xs:double"
XS_INT = "xs:int"
RANGE_STRINGS = ["Range", "range", "RANGE"]
VALUE_TYPE_LOOKUP = {VALUE: "Value", FORMAT: "Format", RANGE: "Range", LIST: "List",}
AGGREGATION = 1
INHERITANCE = 2
SHARED_PARAMETERS = "SharedParameters"
EXISTS = "Exists"
DATA_TYPES = [XS_STRING,XS_LONG,XS_INT,XS_DOUBLE,XS_BOOL]



from __future__ import annotations

import json
import os
from src.SOMcreator import classes
from src.SOMcreator.constants import json_constants, value_constants


def _iter_attributes(property_set: classes.PropertySet, pset_dict: dict) -> None:
    for attribute in property_set.attributes:
        pset_dict[attribute.name] = dict()
        attribute_dict = pset_dict[attribute.name]

        attribute_dict[json_constants.DATA_TYPE] = attribute.data_type
        if not attribute.value:
            attribute_dict[json_constants.VALUE_TYPE] = value_constants.EXISTS
        else:
            attribute_dict[json_constants.VALUE_TYPE] = attribute.value_type
        attribute_dict[json_constants.VALUE] = attribute.value


def export(project: classes.Project, path: str | os.PathLike) -> None:
    json_dict = dict()
    for obj in sorted(project.objects, key=lambda x: x.ident_value):
        if not obj.property_sets:
            continue
        if obj.ident_value is None:
            continue
        json_dict[obj.ident_value] = dict()
        obj_dict = json_dict[obj.ident_value]
        for property_set in obj.property_sets:
            if not property_set.attributes:
                continue
            obj_dict[property_set.name] = dict()
            pset_dict = obj_dict[property_set.name]
            _iter_attributes(property_set, pset_dict)
    with open(path, "w") as file:
        json.dump(json_dict, file, indent=1)



from __future__ import annotations
import csv
from src.SOMcreator import classes


def get_distinct_attributes(property_sets: list[classes.PropertySet]):
    attribute_names = list()

    for property_set in property_sets:
        attribute: classes.Attribute
        attribute_names += [attribute.name for attribute in property_set.attributes]

    distinct_attribute_names = list(dict.fromkeys(attribute_names))

    return distinct_attribute_names


def export_boq(project: classes.Project, path: str, pset_name: str) -> None:
    if not path:
        return

    with open(path, "w", ) as file:
        writer = csv.writer(file, delimiter=";")
        property_sets = [property_set for property_set in classes.PropertySet if
                         property_set.name == pset_name]
        distinct_attribute_names = get_distinct_attributes(property_sets)
        header = ["Ident", "Object"] + [f"{pset_name}:{name}" for name in distinct_attribute_names]
        writer.writerow(header)

        for obj in project.objects:
            if pset_name not in [pset.name for pset in obj.property_sets]:
                continue

            property_set = obj.get_property_set_by_name(pset_name)
            ident = obj.ident_attrib
            line = [f"{ident.property_set.name}:{ident.name}", ident.value[0]]

            for attribute_name in distinct_attribute_names:
                attribute: classes.Attribute = property_set.get_attribute_by_name(attribute_name)

                if attribute is not None:
                    line.append("|".join(attribute.value))
                else:
                    line.append("")
            writer.writerow(line)
